/* --- [ usi_logger.c ] -------------------------------------------------------
* Datum       		  : 09.06.2008
*							 Fehler beseitigt, der 0x3e im Datenstrom  als TWI-STOP interpretiert
* AppNote           : AVR312 - Using the USI module as a I2C slave
*                     benutzt die Fleury-LCD-Bibliothek sowie 
*                             die usi-slave.c von Frank Jonischkies
* Compiler			  : avr-gcc 4.1.2
*		
* Description       : lauscht passiv am TWI-Bus, zeichnet den Datenverkehr mit
*							 und protokolliert wahlweise auf LCD oder COM.
*                     Bei gesetztem Jumper an PortB.4 wird auf LCD ausgegeben,
* 							 ohne Jumper (wenn durch Pullup auf High gezogen) auf LCD
*							 Im Ringpuffer von 40 Byte knnen 20 Datenbyte aufgezeichnet werden 
*                     (beim Vergrern des Arrays auf 44 Byte ergeben sich Fehler in den Daten)
* Tabwidth          : 3
*
****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "usi_logger_slave.h"
#include "usi_logger_lcd.h"

#define STARTMSG  "TWI-Logger "
#define STARTMSG1 ">> LCD"
#define STARTMSG2 ">> @ 38400"
#define BAUD  38400
#define TEILER ((F_CPU /(BAUD * 16l)) -1) 
#define CALC_UBBR(F_CPU, BAUD) ((F_CPU /(BAUD * 16)) -1)
#define LCD 1
#define COM 0

volatile uint8_t TWI_Rx_Head = 0;			// Index wird global vereinbart
volatile uint8_t TWI_Rx_Tail = 0;	
	
char msg[4];										// nur fr die Zahlenumwandlung -> dreistellig + /0
// diese Variable kann entfallen, wenn keine LCD-Ausgabe bentigt wird
char zeile[22];									// nimmt den Inhalt einer LCD-Zeile auf

uint8_t letztesByte;								// das zuletzt empfangene Byte
uint8_t vorletztesByte;							// das vorletzte empfangene Byte
uint8_t Stop_Flag 	= 0;						// Flag 
uint8_t counter 		= 0;						// Zeilenzhler fr die Ausgabe
uint8_t output			= 0;						// 0 = COM, 1 = LCD

void print_on_lcd(void);						// Ausgabe der Steuerzeichen auf LCD
void print_on_com(void);						// Ausgabe der Steuerzeichen auf COM
void print_data(void);							// Ausgabe der Daten, auf LCD und COM

// ----------------------------------------------------------------------------
// ein einzelnes Zeichen auf COM ausgeben
void putc(uint8_t Zeichen)
{
while(!(UCSRA &(1<<UDRE)));
UDR = Zeichen;
}

// ----------------------------------------------------------------------------
// einen String auf COM ausgeben
void putstring(char *s)
{
while(*s)
	{
	putc(*s);
	s++;
	}
}

// ----------------------------------------------------------------------------
// gibt einen String aus dem Flash aus
void puts_p(const char *progmem_s)
{
register char c;
while ( (c = pgm_read_byte(progmem_s++)) )
	{
   putc(c);
	}
}


// ----------------------------------------------------------------------------
int main( void )
{
// alle Variablen sind global deklariert

PORTB = (1<<PB4)	;									// PORTB.4 als Input mit gesetztem Pullup
															// zur Abfrage des Output-Modus
															// nicht sofort den PIN abfragen, sondern etwas warten ...
														
usi_twi_slave_init();								// ohne bergabe einer Adresse !!!
sei();	 												// enable_interrupt

if (PINB & (1<<PB4))									// PINB.4 entscheidet, ob die Ausgabe auf LCD oder COM erfolgt
	{														// wenn high (kein Jumper zieht den Pin auf Masse !)
//	output = COM;										// kann man sich auch sparen ...

	UBRRL  = TEILER;									// serielle Schnittstelle initialisieren
	UCSRB |= (1<<TXEN);
	UCSRC |= (0<<UMSEL) | (1<<UCSZ1) | (1<<UCSZ0);
															// und eine Meldung ausgeben
	putc(13);											// Zeilenumbruch
	puts_p(PSTR(STARTMSG));							// Startmeldung									
	puts_p(PSTR(STARTMSG2));						// Startmeldung vervollstndigen
	putc(13);											// Zeilenumbruch	

	}
else														// wenn Jumper an PINB.4 gesetzt, dann Ausgabe auf LCD
	{
	output = LCD;
	lcd_init(LCD_DISP_ON);							// initialize display, cursor off

	lcd_puts_p(PSTR(STARTMSG));					// Einschaltmeldung auf LCD
	lcd_puts_p(PSTR(STARTMSG1));					// Startmeldung vervollstndigen
	}

Stop_Flag = TRUE;										// Ausgabe solange unterbinden, bis ein START empfangen wird

while(1)
	{

	if (!(TWI_Rx_Tail == TWI_Rx_Head))			// testen, ob ein Byte empfangen
		{
		letztesByte = usi_twi_get_byte();		// ein Byte abholen

															// TWI_RxTail ist der aktuelle Pointer auf das Empfangsarray
		if (!(TWI_Rx_Tail & 0x01))					// in den 'geraden' Stellen steht START, ACK und NACK
			{												// Ausgabe auf LCD bzw. COM
			if (!(letztesByte == STOP))			// ein STOP hier nicht ausgeben, sondern warten bis zum nchsten Byte
				{
				if (output)	print_on_lcd(); else	print_on_com();	
				}
			}
		
		else												// an ungerader Stelle stehen immer die Daten,
			{												// Ausnahme:
			if (vorletztesByte == STOP)			// falls das letzte Steuerzeichen ein STOP war
				{											// dann folgt an der Datenposition das vorausgegangene N/ACK
				if (output)								// Ausgabe auf LCD
					{
					print_on_lcd();					// hier wird das letzteByte (N/ACK) ausgegeben
					letztesByte = vorletztesByte;	// nun noch das STOP umkopieren und
					print_on_lcd();					// STOP ausgeben
					}

				else
					{
					putc(letztesByte);				// das letzteByte enthlt N/ACK
					putc(vorletztesByte);			// das vorletzteByte ist das STOP
					Stop_Flag = TRUE;		
					}
				Stop_Flag = TRUE;
				}

			else
				{
				print_data();							// Daten ausgeben
				}
			}  

		vorletztesByte = letztesByte;				// das letzte Byte kopieren fr den Test auf STOP
		}
	}
}	
		

// ----------------------------------------------------------------------------
// gibt START, ACK, NACK, STOP auf dem LCD aus und kopiert die Ausgabe in Zeilenpuffer
void print_on_lcd(void)
{
if (Stop_Flag && !(letztesByte == START)) return;

switch(letztesByte)
	{
	case START:
		Stop_Flag = FALSE;							// letztes Stop_Flag lschen und 'vorletztesByte' ignorieren
		lcd_clrscr();									// LCD lschen
		lcd_puts(zeile);								// die letzte Zeile2 als Zeile1 ausgeben
		zeile[0] = 0;       							// den String durch Setzen einer 0 lschen
					
		lcd_goto(ZEILE2 + 1);						// an den Start von Zeile 2 gehen
				
													
		if (++counter > 9) counter = 0;			// counter auf LCD nur 2stellig werden lassen !
													
		lcd_puts(itoa(counter, msg, 10));		// den counter ausgeben
		strcat(zeile, msg);							// und in den Zeilenpuffer kopieren

		lcd_data(' ');									// ein Leerzeichen ausgeben
		lcd_data(START);								// das Zeichen ausgeben und dann an Zeilenpuffer anhngen
		strcat(zeile, " <");							// hier ist keine berprfung der Zeilenlnge erforderlich
		break;

	case ACK:
		if (strlen(zeile) < 20)						// die Zeilenlnge des LCD's ist leider beschrnkt
			{
			lcd_data(ACK);
			strcat(zeile, ",");
			}
		break;

	case NACK:
		if (strlen(zeile) < 20) 
			{
			lcd_data(NACK);
			strcat(zeile, ";");
			}
		break;
	
	case STOP:
		Stop_Flag = TRUE;      						// damit zwischen STOP und START keine Flldaten ausgegeben werden

		if (strlen(zeile) < 20) 
			{
			lcd_data(STOP);
			strcat(zeile, ">");
			}

	default:
		break;
	}
}


// ----------------------------------------------------------------------------
// gibt START, ACK, NACK, STOP auf COM aus
void print_on_com(void)
{
// wenn das STOP_FLAG gesetzt, dann wird ausser einem START alles ignoriert
if (Stop_Flag && !(letztesByte == START)) return;

switch(letztesByte)
	{
	case START:
		Stop_Flag = FALSE;
		putc(13);										// Zeilenumbruch
					
		counter++;										// den counter nur 2stellig werden lassen !
		if (counter > 99) counter = 1;
		if (counter < 10) 							// eine fhrende Null ausgeben
			{
			putc('0');	
			} 
		putstring(itoa(counter, msg, 10));
		putc(' ');
		putc(START);
		break;

	case STOP:
		Stop_Flag = TRUE;								// damit zwischen STOP und START keine Flldaten ausgegeben werden
	case ACK:
	case NACK:
		putc(letztesByte);
		break;		
	default:
			putc('?');
		break;
	}
}

// ----------------------------------------------------------------------------
// gibt ein Byte Hexadezimal auf COM oder LCD aus (und hngt es an die Variable 'zeile' an)
void print_data(void)
{

if (Stop_Flag) return;								// keine Ausgabe, wenn das Stop_Flag gesetzt ist

if (letztesByte < 16)								// eine fhrende Null ausgeben (Hex-Darstellung !)
	{
	if (output)											// wenn Ausgabe auf LCD
		{
		if (strlen(zeile) < 20)
			{
			lcd_data('0');
			strcat(zeile, "0");
			}
		}

	else													// wenn Ausgabe auf COM
		{
		putc('0');
		}
	} 
   
itoa(letztesByte, msg, 16);						// letztesByte in hexadezimale Zeichenkette umwandeln
															// die Zeichenfolge steht in msg
															
if (output)												// wenn Ausgabe auf LCD
	{
	lcd_puts(msg);										// letztesByte auf LCD ausgeben
															// letztesByte an die Variable 'zeile' anhngen
	if (letztesByte < 16)							// auch das letzte der 20 LCD-Zeichen ausnutzen
		{													// dann wird nur ein Zeichen geschrieben
		if (strlen(zeile) < 21) strcat(zeile, msg);
		}
	else													// dann werden zwei Zeichen geschrieben
		{
		if (strlen(zeile) < 20) strcat(zeile, msg);
		}
	}

else														// wenn Ausgabe auf COM
	{
	putstring(msg);									// letztesByte hexadezimal auf COM ausgeben
	}
}
