/* --- [ usi_logger.c ] -------------------------------------------------------
* Datum       		  : 09.06.2008
*							 Fehler beseitigt, der 0x3e im Datenstrom als TWI-STOP interpretiert
* AppNote           : AVR312 - Using the USI module as a I2C slave
*                     benutzt die Fleury-LCD-Bibliothek sowie 
*                             die usi-slave.c von Frank Jonischkies
* Compiler			  : avr-gcc 4.1.2
*		
* Description       : lauscht passiv am TWI-Bus, zeichnet den Datenverkehr mit
*							 und protokolliert auf COM.
*							 Im Ringpuffer von 64 Byte knnen 32 Datenbyte aufgezeichnet werden 
*							 
* Tabwidth          : 3
*
****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "usi_logger_slave.h"

#define STARTMSG  "USI-Logger @ 38400"
#define BAUD  38400
#define TEILER ((F_CPU /(BAUD * 16l)) -1) 
#define CALC_UBBR(F_CPU, BAUD) ((F_CPU /(BAUD * 16)) -1)

volatile uint8_t TWI_Rx_Head = 0;			// Index wird global vereinbart
volatile uint8_t TWI_Rx_Tail = 0;	
	
char msg[4];										// nur fr die Zahlenumwandlung -> dreistellig + /0

uint8_t letztesByte;								// das zuletzt   empfangene Byte
uint8_t vorletztesByte;							// das vorletzte empfangene Byte
uint8_t Stop_Flag 	= 0;						// Flag 
uint8_t counter 		= 0;						// Zeilenzhler fr die Ausgabe

void show_control(void);						// Ausgabe der Steuerzeichen START,ACK, NACK, STOP
void show_data(void);							// Ausgabe der bertragenen Datenbytes

// ----------------------------------------------------------------------------
// ein einzelnes Zeichen auf COM ausgeben
void putc(uint8_t Zeichen)
{
while(!(UCSRA &(1<<UDRE)));
UDR = Zeichen;
}

// ----------------------------------------------------------------------------
// einen String auf COM ausgeben
void putstring(char *s)
{
while(*s)
	{
	putc(*s);
	s++;
	}
}

// ----------------------------------------------------------------------------
// gibt einen String aus dem Flash aus
void puts_p(const char *progmem_s)
{
register char c;
while ( (c = pgm_read_byte(progmem_s++)) )
	{
   putc(c);
	}
}

// ----------------------------------------------------------------------------
int main( void )
{
// alle Variablen sind global deklariert

usi_twi_slave_init();								// ohne bergabe einer Adresse !!!
sei();	 												// enable_interrupt

UBRRL  = TEILER;										// serielle Schnittstelle initialisieren
UCSRB |= (1<<TXEN);
UCSRC |= (0<<UMSEL) | (1<<UCSZ1) | (1<<UCSZ0);
															// und eine Meldung ausgeben
putc(13);												// Zeilenumbruch									
									
puts_p(PSTR(STARTMSG));								// Startmeldung
putc(13);												// Zeilenumbruch	

Stop_Flag = TRUE;										// Ausgabe solange unterbinden, bis ein START empfangen wird

while(1)
	{

	if (!(TWI_Rx_Tail == TWI_Rx_Head))			// testen, ob ein Byte empfangen
		{
		letztesByte = usi_twi_get_byte();		// ein Byte abholen
		
															// TWI_RxTail ist der aktuelle Pointer auf das Empfangsarray
		if (!(TWI_Rx_Tail & 0x01))					// in den 'geraden' Stellen steht START, ACK und NACK
			{												// aber Sonderbehandlung bei einem STOP
															// da folgt nmlich das vorausgegangene N/ACK an der nchsten Datenposition
															// daher hier erst einmal garnichts tun, sondern aufs nchste Byte warten
			if (!(letztesByte == STOP)) show_control();	
			}
		
		else												// an ungerader Stelle stehen immer die Daten,
			{									
			if (vorletztesByte == STOP)			// falls jedoch das letzte Steuerzeichen ein STOP war
															// dann folgt an der Datenposition das vorausgegangene N/ACK
				{	
				putc(letztesByte);					// das letzteByte enthlt N/ACK
				putc(vorletztesByte);				// das vorletzteByte ist das STOP
				Stop_Flag = TRUE;
				}

			else
				{
				show_data();							// Daten ausgeben
				}
			}  

		vorletztesByte = letztesByte;				// das letzte Byte kopieren fr den Test auf STOP
		}
	}
}	


// ----------------------------------------------------------------------------
// gibt START, ACK, NACK aus (STOP wird direkt behandelt)
void show_control(void)
{
// wenn das STOP_FLAG gesetzt, dann wird ausser einem START alles ignoriert
if (Stop_Flag && !(letztesByte == START)) return;

switch(letztesByte)
	{
	case START:
		Stop_Flag = FALSE;
		putc(13);										// Zeilenumbruch
					
		counter++;										// den counter nur 2stellig werden lassen !
		if (counter > 99) counter = 1;
		if (counter < 10) 							// eine fhrende Null ausgeben
			{
			putc('0');	
			} 
		putstring(itoa(counter, msg, 10));
		putc(' ');
		putc(START);
		break;

	case STOP:
		Stop_Flag = TRUE;								// damit zwischen STOP und START keine Flldaten ausgegeben werden
	case ACK:
	case NACK:
		putc(letztesByte);
		break;		
	
	default:
		putc('?');
		break;
	}
}

// ----------------------------------------------------------------------------
// gibt ein Byte Hexadezimal auf COM aus 
void show_data(void)
{

if (Stop_Flag) return;								// keine Ausgabe, wenn das Stop_Flag gesetzt ist

if (letztesByte < 16)								// eine fhrende Null ausgeben (Hex-Darstellung !)
	{
	putc('0');
	} 
   
itoa(letztesByte, msg, 16);						// letztesByte in hexadezimale Zeichenkette umwandeln
putstring(msg);										// letztesByte hexadezimal auf COM ausgeben

}
