/*****************************************************************************
*
* Atmel Corporation
*
* File              : usi_logger_slave.c (usi_twi_slave.c)
* Compiler          : WINAVR
* Revision          : 
* Date              : 09.06.08
*							 Fehler beseitigt, der 0x3e im Datenstrom als TWI-STOP interpretiert
* Updated by        : $Author: ltwa $
*
* Support mail      : avr@atmel.com
*
* Supported devices : All device with USI module can be used.
*                     The example is written for the ATmega169, ATtiny26 & ATtiny2313
*
* AppNote           : AVR312 - Using the USI module as a TWI slave
*
* Description       : Functions for USI_TWI_receiver and USI_TWI_transmitter.
*
* Changes for WinAVR by Frank Jonischkies, fjonisch@gmx.de, 11.08.2005
* angepasste/abgespeckte Version fr den Datenlogger
* Tabwidth			  : 3
*
****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include "usi_logger_slave.h"

extern volatile uint8_t TWI_Rx_Tail;		// das volatile ist wichtig !
extern volatile uint8_t TWI_Rx_Head;		// das volatile ist wichtig !
uint8_t TWI_Rx_Buf[TWI_RX_BUFFER_SIZE ];

//-----------------------------------------------------------------------------
// Initialise USI for TWI Slave mode
// Bentigt keine Adresse, da nur (mit-) gelesen wird

void usi_twi_slave_init(void)
{

PORT_USI |=	(1<<PORT_USI_SCL);				// Set SCL high
PORT_USI |=	(1<<PORT_USI_SDA);				// Set SDA high
DDR_USI	|=	(1<<PORT_USI_SCL);				// Set SCL as output
DDR_USI &= ~(1<<PORT_USI_SDA);				// Set SDA as input
USICR 	=	(1<<USISIE)|(0<<USIOIE)|		// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)| 		// Set USI in Two-wire mode. No USI Counter overflow prior
														// to first Start Condition (potentail failure)
				(1<<USICS1)|(0<<USICS0)|		// Shift Register Clock Source = External, positive edge
				(0<<USICLK)|(0<<USITC);
USISR = 0xF0;										// Clear all flags and reset overflow counter
}

// ----------------------------------------------------------------------------
// liefert ein Byte zurck, vorher muss geprft werden, ob neue Daten anstehen
// die Prfung: if (!(TWI_Rx_Tail == TWI_Rx_Head)), beides sind public-Variablen
uint8_t usi_twi_get_byte(void)
{
if ((++TWI_Rx_Tail) >= TWI_RX_BUFFER_SIZE) TWI_Rx_Tail = 0;
//TWI_Rx_Tail = ((TWI_Rx_Tail + 1) & TWI_BUFFER_MASK);
  
return(TWI_Rx_Buf[TWI_Rx_Tail]);
}


//-----------------------------------------------------------------------------
//  Detects the USI_TWI Start Condition and intialises the USI
ISR(USI_START_vect)
{

uint8_t register tmp_Rx_Head;						// spart einige Speicheraktionen
tmp_Rx_Head = TWI_Rx_Head;							// volatile in Register zwischenspeichern
															// fr if und else-Zweig
USISR |= (1<<USIPF);									// Zur Sicherheit Stop-Flag lschen
															// Set default starting conditions for new TWI package
DDR_USI &= ~(1<<PORT_USI_SDA);					// Set SDA as input

if (tmp_Rx_Head & 0x01)								// wir sind an 'ungerader' Speicherstelle
	{														// Index incrementieren
															// an ungerader Stelle kann der Index nach Incrementieren berlaufen
	if ((++tmp_Rx_Head) >= TWI_RX_BUFFER_SIZE) tmp_Rx_Head = 0;
	//tmp_Rx_Head = ((tmp_Rx_Head + 1) & TWI_BUFFER_MASK);
	TWI_Rx_Buf[tmp_Rx_Head] = START;				// an gerader Stelle ein 'START' schreiben
	}

else														// wir sind an 'gerader' Speicherstelle	
	{
															// bei Incrementieren des 'geraden' Index keine Prfung auf berlauf ntig
	TWI_Rx_Buf[++tmp_Rx_Head] = STOP;			// an ungerader Stelle im Buffer ein 'STOP' als Fllbyte schreiben
															// beim Incrementieren aus 'ungeradem' Index ist berlauf mglich
	if ((++tmp_Rx_Head) >= TWI_RX_BUFFER_SIZE) tmp_Rx_Head = 0;
	//tmp_Rx_Head = ((tmp_Rx_Head + 1) & TWI_BUFFER_MASK);
	TWI_Rx_Buf[tmp_Rx_Head] = START;				// an gerader Stelle im Buffer ein 'START' schreiben
	}

TWI_Rx_Head = tmp_Rx_Head;							// volatile wieder im RAM ablegen
															// die Verknpfung muss logisch UND (&&) sein nicht bitweise (&)
while ( (PIN_USI & (1<<PORT_USI_SCL)) && !(USISR & (1<<USIPF)) );
															// Wait for SCL to go low to ensure the "Start Condition" has completed.
                                             // If a Stop condition arises then leave the interrupt to prevent waiting forever.																// Enable TWI Coutner overflow and interrupts
USICR=(1<<USISIE)|(1<<USIOIE)|					// Enable Overflow and Start Condition Interrupt.
															// (Keep StartCondInt to detect RESTART)
		(1<<USIWM1)|(1<<USIWM0)| 					// Set USI in Two-wire mode.
		(1<<USICS1)|(0<<USICS0)|(0<<USICLK)| 	// Shift Register Clock Source = External, positive edge
		(0<<USITC);
USISR=(1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|		// Clear flags
		(1<<USIDC) |(0x0<<USICNT0); 				// Set USI to sample 8 bits i.e. count 16 external pin toggles.
}

//-----------------------------------------------------------------------------
// Handels all the comunication. Is disabled only when waiting for new Start Condition.

ISR(USI_OVERFLOW_vect)
{
uint8_t register tmp_Rx_Head;						// 
tmp_Rx_Head = TWI_Rx_Head;							// volatile in Register zwischenspeichern
															// hier kommen wir mit 'geradem' Index an, kein berlauf mglich 
TWI_Rx_Buf[++tmp_Rx_Head] = USIDR;				// an ungerader Stelle Daten speichern
USIDR=0; 												// prepare to receive N/ACK

USISR=(1<<USISIF)|(1<<USIOIF)|(0<<USIPF)|		// Clear all flags, 
		(1<<USIDC) |(0xE<<USICNT0);				// Set USI to shift out 1 bit
															// Index ist 'ungerade', hier ist Prfung auf berlauf ntig
if ((++tmp_Rx_Head) >= TWI_RX_BUFFER_SIZE) tmp_Rx_Head = 0;
//tmp_Rx_Head = ((tmp_Rx_Head + 1) & TWI_BUFFER_MASK);
															// an gerader STelle n/ACK speichern
while ( !(USISR & ( (1<<USISIF)|(1<<USIOIF) )) ); // bis zum Overflow bzw START warten

if (USIDR)												// wenn SDA high (nicht auf Masse gezogen),
	{
	TWI_Rx_Buf[tmp_Rx_Head] = NACK;				// dann ist das ein NACK
	}
else														// wenn SDA low (auf Masse gezogen), 
	{
	TWI_Rx_Buf[tmp_Rx_Head] = ACK;				// dann ist das ein ACK
	}

USISR=(1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|		// Alle flags lschen, die nachfolgend abgefragt werden !!!
		(1<<USIDC) |(0xE<<USICNT0); 				// Set USI to sample 1 bit
															// Wait until 1 bit sampled OR STOP- OR START- condition.

while ( !(USISR & ( (1<<USISIF)|(1<<USIOIF)|(1<<USIPF) )) );

if ( USISR & (1<<USIPF) )							// If STOP Cond
	{														// Set USI in TWI Start Condition Mode
	USICR=(1<<USISIE)|(0<<USIOIE)|				// Enable Start Condition Interrupt. Disable Overflow Interrupt.
			(1<<USIWM1)|(0<<USIWM0)|				// Set USI in Two-wire mode. No USI Counter overflow hold.
			(1<<USICS1)|(0<<USICS0)|(0<<USICLK)|// Shift Register Clock Source = External, positive edge
			(0<<USITC);
	USISR=(0<<USISIF)|(1<<USIOIF)|(1<<USIPF)|	// Clear all flags but Start Condition
			(1<<USIDC) |(0x0<<USICNT0);
															// hier kommen wir mit 'geradem' Index an, keine berprfung
															// aber: das STOP muss an gerader Stelle gespeichert werden, darum das
															// N/ACK an die Position der Daten schieben
	TWI_Rx_Buf[tmp_Rx_Head + 1] = TWI_Rx_Buf[tmp_Rx_Head]; // an ungerader Stelle das vorausgegangene N/ACK ablegen
	TWI_Rx_Buf[tmp_Rx_Head++] = STOP;			// an gerader Stelle STOP speichern und Index incrementieren
	}

else
	{
	USISR=(0<<USISIF)|(1<<USIOIF)|(0<<USIPF)|	// Clear all flags, except Start Cond
			(1<<USIDC) |(0x2<<USICNT0);			// Set USI to sample last 7 bits
	}

TWI_Rx_Head = tmp_Rx_Head;							// incrementierte volatile wieder im RAM ablegen
}
